const fs = require('fs');
const imageType = require('image-type');
const path = require('path');
const readChunk = require('read-chunk');
const sanitize = require('sanitize-filename');
const zip = require('gulp-zip');
const webpack = require('webpack');
const prefixer = require('postcss-prefix-selector'); // używany przez webpacka jako string
const MiniCssExtractPlugin = require('mini-css-extract-plugin');
const TerserPlugin = require('terser-webpack-plugin');

const { distPath, packagePath, manifestPath } = require('./const');

function build(gulp) {
    return async function build() {
        if (!fs.existsSync(distPath)) fs.mkdirSync(distPath);

        const { packageName, solutionId, ...manifest } = getManifest();

        await processSingleEntry('component', solutionId, false);
        await processSingleEntry('emptyValueValidator', solutionId);
        await processSingleEntry('additionalValidator', solutionId);

        copyAssets(manifest);
        copyManifest();

        return gulp
            .src(`${distPath}/**/*`)
            .pipe(zip(`${sanitize(packageName)}.bpspkg`))
            .pipe(gulp.dest(packagePath));
    };
}

function copyAssets(manifest) {
    const assets = manifest.files.assets;

    if (typeof assets !== 'object') {
        return;
    }

    const assetsPath = `${distPath}/assets`;
    const keys = Object.keys(assets);
    const filesToCopy = keys.map(n => ({ key: n, path: assets[n] }));

    if (!fs.existsSync(assetsPath)) fs.mkdirSync(assetsPath);

    for (let i = 0; i < filesToCopy.length; i++) {
        const file = filesToCopy[i];

        if (!fs.existsSync(file.path)) {
            throw Error(`Asset file '${file.key}' does not exist.`);
        }

        const chunk = readChunk.sync(file.path, 0, imageType.minimumBytes);
        const isImage = imageType(chunk);

        if (!isImage) {
            throw Error(`Asset file '${file.key}' is not an image file.`);
        }

        const ext = path.extname(file.path);

        fs.copyFileSync(file.path, `${assetsPath}/${file.key}${ext}`);
    }
}

function copyManifest() {
    fs.copyFileSync(manifestPath, `${distPath}/manifest.json`);
}

function getManifest() {
    const manifestBuffer = fs.readFileSync(manifestPath);
    const manifest = JSON.parse(manifestBuffer);

    return manifest;
}

function getPath(type, optional = true) {
    const manifest = getManifest();

    if (typeof manifest.files !== 'object') {
        throw Error('Invalid manifest file.');
    }

    const path = manifest.files[type];

    if (optional && path === undefined) {
        return null;
    }

    if (path === undefined || !fs.existsSync(path)) {
        throw Error(`File '${path}' for '${type}' file type does not exist.`);
    }

    return path;
}

async function processSingleEntry(type, solutionId, optional = true) {
    const path = getPath(type, optional);
    const solutionIdForCss = solutionId.split('-').join('_');

    if (!path) {
        return;
    }
    const compiler = webpack({
        mode: 'production',
        entry: {
            [solutionId]: path,
        },
        devtool: false,
        optimization: {
            minimize: true,
            minimizer: [
                new TerserPlugin({
                    extractComments: false,
                    terserOptions: {
                        format: {
                            comments: false,
                        },
                    },
                }),
            ],
            splitChunks: {
                cacheGroups: {
                    styles: {
                        name: 'styles',
                        type: 'css/mini-extract',
                        chunks: 'all',
                        enforce: true,
                    },
                },
            },
        },
        plugins: [
            new MiniCssExtractPlugin({
                filename: 'bundle.css',
            }),
        ],
        module: {
            rules: [
                {
                    test: /\.(?:js|mjs|cjs)$/,
                    exclude: /node_modules/,
                    use: {
                        loader: require.resolve('babel-loader'),
                        options: {
                            presets: [
                                [
                                    require.resolve('@babel/preset-env'),
                                    { targets: 'defaults' },
                                ],
                                [
                                    require.resolve('@babel/preset-react'),
                                    { pragma: 'webcon.createElement' },
                                ],
                            ],
                        },
                    },
                },
                {
                    test: /\.css$/i,
                    use: [
                        {
                            loader: MiniCssExtractPlugin.loader,
                        },
                        { loader: require.resolve('css-loader') },
                        {
                            loader: require.resolve('postcss-loader'),
                            options: {
                                postcssOptions: {
                                    plugins: {
                                        'postcss-prefix-selector': {
                                            prefix: `.dynamic-form .component_${solutionIdForCss}`,
                                            transform(
                                                prefix,
                                                selector,
                                                prefixedSelector,
                                                filePath,
                                                rule
                                            ) {
                                                if (
                                                    selector.match(
                                                        /^(html|body)/
                                                    )
                                                ) {
                                                    return selector.replace(
                                                        /^([^\s]*)/,
                                                        `$1 ${prefix}`
                                                    );
                                                }

                                                if (
                                                    filePath.match(
                                                        /node_modules/
                                                    )
                                                ) {
                                                    return selector; // Do not prefix styles imported from node_modules
                                                }

                                                const annotation = rule.prev();
                                                if (
                                                    annotation?.type ===
                                                        'comment' &&
                                                    annotation.text.trim() ===
                                                        'no-prefix'
                                                ) {
                                                    return selector; // Do not prefix style rules that are preceded by: /* no-prefix */
                                                }

                                                return prefixedSelector;
                                            },
                                        },
                                    },
                                },
                            },
                        },
                    ],
                },
            ],
        },
        externals: {
            react: 'React',
            ['react-dom']: 'ReactDOM',
        },
        output: {
            filename: `${type}.js`,
            library: ['webcon', 'customControls', `${type}s`, solutionId],
            libraryTarget: 'window',
            libraryExport: 'default',

            environment: {
                arrowFunction: false,
            },
        },
    });

    await new Promise((resolve, reject) => {
        compiler.run((err, res) => {
            if (err) {
                return reject(err);
            }
            resolve(res);
        });
    });
}

exports.build = build;
